import numpy as np
import qcelemental as qcel
import os


def run_checkmol(mol_file, checkmol_path='checkmol'):
    """Runs Checkmol on a given MOL file and returns the output file path."""
    mol_base_path = mol_file.split(".")[0]
    filename = mol_file.split("/")[-1]
    checkmol_file = mol_base_path + ".checkmol"
    os.system(f"{checkmol_path} -vxpe {mol_base_path + filename} > {checkmol_file}")
    return checkmol_file


def compute_closest_hydrogens(active_ind_distances, H_atom_indices, Htot):
    """Computes the closest H_atoms to a given atom index based on the distance matrix."""
    H_distances = active_ind_distances[H_atom_indices]
    closest_H_indices = np.argsort(H_distances)[:int(Htot)]
    return H_atom_indices[closest_H_indices]


def process_checkmol_data(checkmol_file, verbose=False):
    """
    Processes the output of a Checkmol file and returns a dictionary with the following keys:
    - atomic_numbers
    - atomic_types
    - aromatic
    - heavy_atom_neighbors
    - Hexps
    - Htots
    - functional_group_indices
    - functional_group_labels
    """
    with open(checkmol_file, "r") as f:
        lines = f.readlines()
    if verbose:
        print("".join(lines))
    xyz = []
    atomic_numbers = []
    atomic_types = []
    aromatic_label = []
    functional_group_labels = []
    functional_group_indices = []
    heavy_atom_neighbors = []
    Hexps = []
    Htots = []
    H_atom_indices = []
    for line in lines:
        if "heavy-atom" in line:
            line = line.split()
            ind, atom, atom_type, x, y, z, hatom_neighbors, Hexp, Htot = line[0], line[1], line[2], line[3], line[4], line[5], line[6], line[10], line[12]
            aromatic = "aromatic" in line
            xyz.append([float(x), float(y), float(z)])
            atomic_numbers.append(qcel.periodictable.to_Z(atom))
            atomic_types.append(atom_type)
            aromatic_label.append(aromatic)
            heavy_atom_neighbors.append(int(hatom_neighbors.replace("(", "")))
            Hexps.append(int(Hexp))
            Htots.append(int(Htot.replace(")", "")))
            if "H" in atom:
                H_atom_indices.append(int(ind) - 1)
        elif line.startswith("#"):
            split_line = line.split(":")
            functional_group_label = ":".join(split_line[0:2])
            indices = ":".join(split_line[2:]).strip()
            for n, p in enumerate(indices.split(",")):
                output_indices = []
                explicit_indices = p.split("-")
                if explicit_indices[-1] == '':
                    explicit_indices = explicit_indices[:-1]
                # checkmol uses 1-indexing but need 0-indexing for np
                for i in explicit_indices:
                    if ":" in i:
                        start, end = i.split(":")
                        output_indices += list(range(int(start) - 1, int(end)))
                    else:
                        output_indices.append(int(i) - 1)
                if n > 0:
                    # check to see if any output_indices overlap with previous output_indices, if so merge
                    if np.any(np.isin(output_indices, functional_group_indices[-1])):
                        functional_group_indices[-1] = np.concatenate((functional_group_indices[-1], output_indices))
                        continue
                functional_group_indices.append(output_indices)
                functional_group_labels.append(functional_group_label)
    H_atom_indices = np.array(H_atom_indices)
    xyz = np.array(xyz)
    atomic_numbers = np.array(atomic_numbers)
    distance_matrix = np.linalg.norm(xyz[:, None] - xyz, axis=-1)
    # H_distances = distance_matrix[H_atom_indices]
    # Now we have all information to identify which hydrogens belong to certain
    # functional groups. For every non-zero Htots element, we will use Htot
    # index to identify the closest H_atoms to the current atom. Then grab the
    # Htot number of closest H_atoms indices to append to the functional group
    # indices.
    for n, f_indices in enumerate(functional_group_indices):
        if verbose:
            print(n, f_indices)
            print(f"Functional group: {functional_group_labels[n]}")
            print(f"Functional group indices: {functional_group_indices[n]}")
        for i in f_indices:
            Htot = Htots[i]
            if Htot != 0:
                ind_distances = distance_matrix[i]
                closest_H_indices = compute_closest_hydrogens(ind_distances, H_atom_indices, Htot)
                functional_group_indices[n] = np.concatenate((functional_group_indices[n], closest_H_indices))
                functional_group_indices[n] = np.unique(functional_group_indices[n])
                if verbose:
                    print("Active index:", i, "Htot:", Htot)
                    print(closest_H_indices)
        if verbose:
            print(f"Functional group indices: {functional_group_indices[n]}")
    return {
        "atomic_numbers": atomic_numbers,
        "atomic_types": atomic_types,
        "aromatic": aromatic_label,
        "heavy_atom_neighbors": heavy_atom_neighbors,
        "Hexps": Hexps,
        "Htots": Htots,
        "functional_group_indices": functional_group_indices,
        "functional_group_labels": functional_group_labels,
    }



import os
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

kcal_per_mol = "$\mathrm{kcal\cdot mol^{-1}}$"


def create_minor_y_ticks(ylim):
    diff = abs(ylim[1] - ylim[0])
    if diff > 100:
        inc = 10
    if diff > 20:
        inc = 5
    elif diff > 10:
        inc = 2.5
    else:
        inc = 1
    lower_bound = int(ylim[0])
    while lower_bound % inc != 0:
        lower_bound -= 1
    upper_bound = int(ylim[1])
    while upper_bound % inc != 0:
        upper_bound += 1
    upper_bound += inc
    minor_yticks = np.arange(lower_bound, upper_bound, inc)
    return minor_yticks


def violin_plot(
    df,
    df_labels_and_columns: {},
    output_filename: str,
    plt_title: str = None,
    output_path: str = "./",
    bottom: float = 0.4,
    ylim: list = None,
    transparent: bool = False,
    widths: float = 0.85,
    figure_size: tuple = None,
    set_xlable=False,
    x_label_rotation=90,
    x_label_fontsize=8,
    ylabel=r"Error ($\mathrm{kcal\cdot mol^{-1}}$)",
    dpi=600,
    usetex=True,
    rcParams={
        "text.usetex": True,
        "font.family": "sans-serif",
        "font.sans-serif": "Helvetica",
        "mathtext.fontset": "custom",
    },
    colors: list = None,
) -> None:
    """
    Create a dataframe with columns of errors pre-computed for generating
    violin plots with MAE, RMSE, and MaxAE displayed above each violin.

    Args:
        df: DataFrame with columns of errors 
        df_labels_and_columns: Dictionary of plotted labels along with the df column for data
        output_filename: Name of the output file
        ylim: list =[-15, 35],
        rcParams: can be set to None if latex is not used
        colors: list of colors for each df column plotted. A default will alternate between blue and green.
    """
    print(f"Plotting {output_filename}")
    if rcParams is not None:
        plt.rcParams.update(rcParams)
    vLabels, vData = [], []
    annotations = []  # [(x, y, text), ...]
    cnt = 1
    plt.rcParams["text.usetex"] = usetex
    for k, v in df_labels_and_columns.items():
        df[v] = pd.to_numeric(df[v])
        df_sub = df[df[v].notna()].copy()
        vData.append(df_sub[v].to_list())
        k_label = "\\textbf{" + k + "}"
        vLabels.append(k_label)
        m = df_sub[v].max()
        rmse = df_sub[v].apply(lambda x: x**2).mean() ** 0.5
        mae = df_sub[v].apply(lambda x: abs(x)).mean()
        max_error = df_sub[v].apply(lambda x: abs(x)).max()
        text = r"\textit{%.2f}" % mae
        text += "\n"
        text += r"\textbf{%.2f}" % rmse
        text += "\n"
        text += r"\textrm{%.2f}" % max_error
        annotations.append((cnt, m, text))
        cnt += 1

    pd.set_option("display.max_columns", None)
    fig = plt.figure(dpi=dpi)
    if figure_size is not None:
        plt.figure(figsize=figure_size)
    ax = plt.subplot(111)
    vplot = ax.violinplot(
        vData,
        showmeans=True,
        showmedians=False,
        quantiles=[[0.05, 0.95] for i in range(len(vData))],
        widths=widths,
    )
    for n, partname in enumerate(["cbars", "cmins", "cmaxes", "cmeans"]):
        vp = vplot[partname]
        vp.set_edgecolor("black")
        vp.set_linewidth(1)
        vp.set_alpha(1)
    quantile_color = "red"
    quantile_style = "-"
    quantile_linewidth = 0.8
    for n, partname in enumerate(["cquantiles"]):
        vp = vplot[partname]
        vp.set_edgecolor(quantile_color)
        vp.set_linewidth(quantile_linewidth)
        vp.set_linestyle(quantile_style)
        vp.set_alpha(1)

    colors = ["blue" if i % 2 == 0 else "green" for i in range(len(vLabels))]
    for n, pc in enumerate(vplot["bodies"], 1):
        pc.set_facecolor(colors[n - 1])
        pc.set_alpha(0.6)

    vLabels.insert(0, "")
    xs = [i for i in range(len(vLabels))]
    xs_error = [i for i in range(-1, len(vLabels) + 1)]
    ax.plot(
        xs_error,
        [1 for i in range(len(xs_error))],
        "k--",
        label=r"$\pm$1 $\mathrm{kcal\cdot mol^{-1}}$",
        zorder=0,
        linewidth=0.6,
    )
    ax.plot(
        xs_error,
        [0 for i in range(len(xs_error))],
        "k--",
        linewidth=0.5,
        alpha=0.5,
        # label=r"Reference Energy",
        zorder=0,
    )
    ax.plot(
        xs_error,
        [-1 for i in range(len(xs_error))],
        "k--",
        zorder=0,
        linewidth=0.6,
    )
    ax.plot(
        [],
        [],
        linestyle=quantile_style,
        color=quantile_color,
        linewidth=quantile_linewidth,
        label=r"5-95th Percentile",
    )
    navy_blue = (0.0, 0.32, 0.96)
    ax.set_xticks(xs)
    plt.setp(
        ax.set_xticklabels(vLabels),
        rotation=x_label_rotation,
        fontsize=x_label_fontsize,
    )
    ax.set_xlim((0, len(vLabels)))
    if ylim is not None:
        ax.set_ylim(ylim)
        minor_yticks = create_minor_y_ticks(ylim)
        ax.set_yticks(minor_yticks, minor=True)

    lg = ax.legend(loc="upper left", edgecolor="black", fontsize="8")

    if set_xlable:
        ax.set_xlabel("Level of Theory", color="k")
    ax.set_ylabel(ylabel, color="k")

    ax.grid(color="#54585A", which="major", linewidth=0.5, alpha=0.5, axis="y")
    ax.grid(color="#54585A", which="minor", linewidth=0.5, alpha=0.5)
    # Annotations of RMSE
    for x, y, text in annotations:
        ax.annotate(
            text,
            xy=(x, y),
            xytext=(x, y + 0.1),
            color="black",
            fontsize="8",
            horizontalalignment="center",
            verticalalignment="bottom",
        )

    for n, xtick in enumerate(ax.get_xticklabels()):
        xtick.set_color(colors[n - 1])
        xtick.set_alpha(0.8)

    if plt_title is not None:
        plt.title(f"{plt_title}")
    fig.subplots_adjust(bottom=bottom)
    ext = "png"
    if len(output_filename.split(".")) > 1:
        ext = output_filename.split(".")[-1]
    output_basename = os.path.basename(output_filename)
    plt.savefig(
        f"{output_path}/{output_basename}_violin.{ext}",
        transparent=transparent,
        bbox_inches="tight",
        dpi=dpi,
    )
    plt.clf()
    return


if __name__ == "__main__":
    # Fake data generated for example
    n_samples = 1000
    mean = 0.5  # replace with your desired mean
    std_dev = 5  # replace with your desired standard deviation
    df = pd.DataFrame(
        {
            "MP2": std_dev * np.random.randn(n_samples) + mean,
            "HF": std_dev * np.random.randn(n_samples) - mean,
            "MP2.5": std_dev * np.random.randn(n_samples) + mean,
        }
    )
    # Only specify columns you want to plot
    vals = {
        "MP2 label": "MP2",
        "HF label": "HF",
    }
    violin_plot(df, vals, ylim=[-20, 35], output_filename="example")



import pickle
import numpy as np
from .periodictable import create_pt_dict, create_el_num_to_symbol
import qcelemental as qcel
import json
import subprocess
import os


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def dict_to_json(d: dict, fn: str):
    with open(fn, "w") as f:
        json_dump = json.dumps(d, indent=4, cls=NumpyEncoder)
        f.write(json_dump)
    return


def json_to_dict(fn: str, return_numpy=False):
    if not os.path.exists(fn):
        return None
    with open(fn, "r") as f:
        d = json.load(f)
        for k, v in d.items():
            if type(v) == list:
                d[k] = np.array(v)
    return d


def save_pkl(file_name, obj):
    with open(file_name, "wb") as fobj:
        pickle.dump(obj, fobj)


def load_pkl(file_name):
    with open(file_name, "rb") as fobj:
        return pickle.load(fobj)


def np_carts_to_string(carts):
    w = ""
    for n, r in enumerate(carts):
        e, x, y, z = r
        line = "{:d}\t{:.10f}\t{:.10f}\t{:.10f}".format(int(e), x, y, z)
        if n != len(carts) - 1:
            line += "\n"
        w += line
    return w


def generate_p4input_from_df(
    geometry,
    charges,
    monAs,
    monBs=None,
    units="angstrom",
    extra=None,  # ="symmetry c1\nno_reorient\n no_com"
):
    if isinstance(geometry, list):
        geometry = np.array(geometry)
        monAs = np.array(monAs).flatten()
        if monBs is not None:
            monBs = np.array(monBs).flatten()
    if monBs is not None:
        ma, mb = [], []
        for i in monAs:
            ma.append(geometry[i, :])
        for i in monBs:
            mb.append(geometry[i, :])
        ma = np_carts_to_string(ma)
        mb = np_carts_to_string(mb)
        geom = f"{charges[1][0]} {charges[1][1]}\n{ma}"
        geom += f"\n--\n{charges[2][0]} {charges[2][1]}\n{mb}"
        if units == "angstrom":
            geom += "\nunits angstrom"
        elif units == "bohr":
            geom += "\nunits bohr"
        else:
            raise ValueError("units must be either angstrom or bohr")
    else:
        ma = np_carts_to_string(geometry)
        geom = f"{charges[0]} {charges[1]}\n{ma}"
        if units == "angstrom":
            geom += "\nunits angstrom"
        elif units == "bohr":
            geom += "\nunits bohr"
        else:
            raise ValueError("units must be either angstrom or bohr")
    if extra is not None:
        geom += f"\n{extra}"
    return geom


def generate_mol_from_df_row_tl(d1, units="angstrom"):
    ma = d1["RA"]
    mb = d1["RB"]
    ma = np_carts_to_string(ma)
    mb = np_carts_to_string(mb)
    charges = d1["charges"]
    geom = f"{charges[1][0]} {charges[1][1]}\n{ma}"
    geom += f"\n--\n{charges[2][0]} {charges[2][1]}\n{mb}"
    if units == "angstrom":
        geom += "\nunits angstrom"
    elif units == "bohr":
        geom += "\nunits bohr"
    else:
        raise ValueError("units must be either angstrom or bohr")
    mol = qcel.models.Molecule.from_data(geom)
    # check if mol is nan
    return mol


def convert_schr_row_to_mol(r) -> qcel.models.Molecule:
    """
    convert_schr_row_to_mol
    """
    ma, mb = r["monAs"], r["monBs"]
    g1, g2 = r["Geometry"][ma], r["Geometry"][mb]
    m1 = f"{r['charges'][1][0]} {r['charges'][1][1]}\n"
    m1 += print_cartesians_pos_carts(g1[:, 0], g1[:, 1:], only_results=True)
    m2 = f"{r['charges'][2][0]} {r['charges'][2][1]}\n"
    m2 += print_cartesians_pos_carts(g2[:, 0], g2[:, 1:], only_results=True)
    mol = qcel.models.Molecule.from_data(m1 + "--\n" + m2)
    return mol


def convert_pos_carts_to_mol(pos, carts):
    m1 = ""
    for i in range(len(pos)):
        if i > 0:
            m1 += "--\n"
        m1 += f"0 1\n"
        m1 += print_cartesians_pos_carts(pos[i], carts[i], only_results=True)
    mol = qcel.models.Molecule.from_data(m1)
    return mol


def string_carts_to_np(geom):
    geom = geom.split("\n")
    if geom[0] == "":
        geom = geom[1:]
    mols = []
    m = []
    charges = [[0, 1]]
    new_mol = False
    el_dict = create_pt_dict()
    monA, monB = [], []
    for n, i in enumerate(geom):
        if n == 0:
            cnt = 0
            m_ind = []
            i = [int(k) for k in i.strip().split(" ")]
            charges.append(i)
        elif new_mol:
            i = [int(k) for k in i.strip().split(" ")]
            charges.append(i)
            new_mol = False
        elif "--" in i:
            new_mol = True
            # mols.append(m)
            monA = m_ind
            m_ind = []
        else:
            i = (
                i.replace("  ", " ")
                .replace("  ", " ")
                .replace("  ", " ")
                .replace("  ", " ")
                .replace("  ", " ")
                .replace("  ", " ")
                .replace("  ", " ")
                .replace("  ", " ")
            ).rstrip()
            i = i.split(" ")
            if i[1].isnumeric():
                el = int(i[1])
            else:
                el = el_dict[i[1]]
            r = [
                el,
                float(i[2]),
                float(i[3]),
                float(i[4]),
            ]
            m.append(np.array(r))
            m_ind.append(cnt)
            cnt += 1
    m = np.array(m)
    monB = m_ind
    charges = np.array(charges)
    monA = np.array(monA)
    monB = np.array(monB)
    return m, charges, monA, monB


def print_cartesians(arr, symbols=False):
    """
    prints a 2-D numpy array in a nicer format
    """
    shape = np.shape(arr)
    if shape[1] != 4:
        raise ValueError("Array must be Nx4")

    l = ""
    for a in arr:
        for i, elem in enumerate(a):
            if i == 0:
                elem = qcel.periodictable.to_E(int(elem))
                print("{} ".format(elem), end="\t")
                l += "{} \t".format(elem)
            else:
                print("{:.10f} ".format(elem).rjust(3), end="\t")
                l += "{:.10f} ".format(elem).rjust(3)
                l += "\t"
        print(end="\n")
        l += "\n"
    return l


def print_cartesians_dimer(geom, monAs, monBs, charges) -> str:
    """
    print_cartesians_dimer takes in dimer geometry and splits
    by monAs and monBs slicing to produce monomers. The
    charges are mult and charge.
    """
    m1 = geom[monAs]
    m2 = geom[monBs]
    c1, c2 = charges[1], charges[2]
    print(*c1)
    print_cartesians(m1)
    print(f"--")
    print(*c2)
    print_cartesians(m2)
    return


def print_cartesians_pos_carts(
    pos: np.array, carts: np.array, only_results=False, el_attach=None
):
    """
    prints a 2-D numpy array in a nicer format
    """
    if not only_results:
        print()
    lines = ""
    for n, r in enumerate(carts):
        x, y, z = r
        el = str(int(pos[n]))
        if el_attach is not None:
            el += el_attach
        line = "{}\t{:.10f}\t{:.10f}\t{:.10f}".format(el, x, y, z)
        lines += line + "\n"
        if not only_results:
            print(line)
    if not only_results:
        print()
    return lines


def print_cartesians_pos_carts_symbols(
    pos: np.array,
    carts: np.array,
    only_results=False,
    el_attach=None,
    el_dc=create_el_num_to_symbol(),
):
    """
    prints a 2-D numpy array in a nicer format
    """
    if not only_results:
        print()
    lines = ""
    for n, r in enumerate(carts):
        x, y, z = r
        el = el_dc[int(pos[n])]
        if el_attach is not None:
            el += el_attach
        line = "{}\t{:.10f}\t{:.10f}\t{:.10f}".format(el, x, y, z)
        lines += line + "\n"
        if not only_results:
            print(line)
    if not only_results:
        print()
    return lines


# def return_cartesians_pos_carts(pos: np.array, carts: np.array):
#     """
#     prints a 2-D numpy array in a nicer format
#     """
#     print()
#     lines = ""
#     for n, r in enumerate(carts):
#         x, y, z = r
#         line = "{}\t{:.10f}\t{:.10f}\t{:.10f}".format(int(pos[n]), x, y, z)
#         lines += line + "\n"
#         print(line)
#     print()
#     return lines


def carts_to_xyz(pos: np.array, carts: np.array, el_dc=create_el_num_to_symbol()):
    """
    creates xyz file from pos and carts
    """
    out = ""
    start = f"{len(pos)}\n\n"
    out += start
    for n, r in enumerate(carts):
        x, y, z = r
        line = "{}\t{:.10f}\t{:.10f}\t{:.10f}\n".format(el_dc[int(pos[n])], x, y, z)
        out += line
    return out


def write_cartesians_to_xyz(
    pos: np.array,
    carts: np.array,
    fn="out.xyz",
    charge_multiplicity=None,
    charge=False,
    multiplicty=False,
):
    """
    creates xyz file from pos and carts
    """
    el_dc = create_el_num_to_symbol()
    out = ""
    with open(fn, "w") as f:
        cm = ""
        if charge_multiplicity is not None and charge and multiplicty:
            cm = f"{charge_multiplicity[0]} {charge_multiplicity[1]}"
        elif charge_multiplicity is not None and charge:
            # Targetting xyz2mol reading charge to RDKit molecule
            cm = f"charge={charge_multiplicity[0]}"
        elif charge_multiplicity is not None and multiplicty:
            cm = f"multiplicity={charge_multiplicity[1]}"
        start = f"{len(pos)}\n{cm}\n"
        out += start
        f.write(start)
        for n, r in enumerate(carts):
            x, y, z = r
            line = "{}\t{:.10f}\t{:.10f}\t{:.10f}\n".format(el_dc[int(pos[n])], x, y, z)
            out += line
            f.write(line)
    return out


def write_pickle(data, fname="data.pickle"):
    with open(fname, "wb") as handle:
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)


def read_pickle(fname="data.pickle"):
    with open(fname, "rb") as handle:
        return pickle.load(handle)


def read_xyz_to_pos_carts(
    xyz_path="mol.xyz",
    array_2d=False,
    charge_mult=False,
) -> (np.array, np.array):
    """
    read_xyz_to_pos_carts reads xyz file and returns pos and carts
    """
    el_dc = create_pt_dict()
    with open(xyz_path, "r") as f:
        data = f.readlines()
        d = data[2:]

    pos, carts = [], []
    for l in d:
        l = l.split()
        if l[0].isnumeric():
            el = int(l[0])
        else:
            el = el_dc[l[0]]
        x, y, z = float(l[1]), float(l[2]), float(l[3])
        pos.append(el)
        carts.append([x, y, z])
    pos = np.array(pos)
    carts = np.array(carts)
    if charge_mult:
        cm = data[1].split()
        cm = [int(cm[0]), int(cm[1])]
    if array_2d and charge_mult:
        return np.hstack((pos.reshape(-1, 1), carts)), cm
    elif array_2d:
        return np.hstack((pos.reshape(-1, 1), carts))
    elif charge_mult:
        return pos, carts, cm

    return pos, carts


def read_geom_to_pos_carts_nmers(
    xyz_path="mol.xyz",
    start=4,
) -> [np.ndarray, np.ndarray]:
    """
    read_xyz_to_pos_carts reads xyz file and returns pos and carts for AP-Net prediction
    """
    el_dc = create_pt_dict()
    with open(xyz_path, "r") as f:
        d = f.readlines()[start:]

    pos, carts = [[]], [[]]
    pc_ind = 0
    for l in d:
        if "--" in l:
            pc_ind += 1
            pos.append([])
            carts.append([])
            continue

        l = l.split()
        if l[0].isnumeric():
            el = int(l[0])
        else:
            try:
                el = el_dc[l[0]]
            except (KeyError, IndexError):
                el = l[0][0] + l[0][1:].lower()
                el = el_dc[el]
                print("Warning: element not in periodic table, using first letter")
                return [], []
        x, y, z = float(l[1]), float(l[2]), float(l[3])
        pos[pc_ind].append(el)
        carts[pc_ind].append([x, y, z])
    for i in range(len(pos)):
        pos[i] = np.array(pos[i])
        carts[i] = np.array(carts[i])
    if len(pos) == 1:
        pos = pos[0]
        carts = carts[0]
    return pos, carts


def convert_geom_str_to_dimer_splits(
    geom, units_angstroms=True
) -> [np.array, np.array, np.array, np.array]:
    """
    convert_str_to_dimer_splits takes in geom as a STRING as a list or single string
    and makes Molecule objects

    returning order [ZA, ZB, RA, RB]
    """
    m = 1
    if units_angstroms:
        m = qcel.constants.conversion_factor("bohr", "angstrom")
    if type(geom) == str:
        mol = qcel.models.Molecule.from_data(geom)
        RA = mol.geometry[mol.fragments[0]] * m
        RB = mol.geometry[mol.fragments[1]] * m
        ZA = mol.atomic_numbers[mol.fragments[0]]
        ZB = mol.atomic_numbers[mol.fragments[1]]
        TQA = mol.fragment_charges[0]
        TQB = mol.fragment_charges[1]
        # MA = mol.fragment_multiplicity[mol.fragments[0]]
        # MB = mol.fragment_multiplicity[mol.fragments[1]]
        return [ZA, ZB, RA, RB, TQA, TQB]
    elif type(geom) == list:
        out = []
        for i in geom:
            mol = qcel.models.Molecule.from_data(i)
            RA = mol.geometry[mol.fragments[0]] * m
            RB = mol.geometry[mol.fragments[1]] * m
            ZA = mol.atomic_numbers[mol.fragments[0]]
            ZB = mol.atomic_numbers[mol.fragments[1]]
            TQA = mol.fragment_charges[0]
            TQB = mol.fragment_charges[1]
            # MA = mol.fragment_multiplicity[mol.fragments[0]]
            # MB = mol.fragment_multiplicity[mol.fragments[1]]
            out.append([ZA, ZB, RA, RB, TQA, TQB])
        return out
    else:
        print("Type not supported")
        return []


def mol_to_pos_carts_ma_mb(mol, units_angstroms=True):
    cD = mol.geometry
    if units_angstroms:
        cD = cD * qcel.constants.conversion_factor("bohr", "angstrom")
    pD = mol.atomic_numbers
    geom = np.hstack((pD.reshape(-1, 1), cD))
    ma = list(mol.fragments[0])
    mb = list(mol.fragments[1])
    charges = np.array(
        [
            [int(mol.molecular_charge), int(mol.molecular_multiplicity)],
            [int(mol.fragment_charges[0]), (mol.fragment_multiplicities[0])],
            [int(mol.fragment_charges[1]), (mol.fragment_multiplicities[1])],
        ]
    )
    qcel.models.Molecule.from_data
    return geom, pD, cD, ma, mb, charges


def mol_qcdb_to_pos_carts_ma_mb(mol, units_angstroms=True):
    p4_input = mol.format_molecule_for_psi4()
    p4_input = "\n".join(p4_input.split("\n")[1:-2])
    geom = mol.format_molecule_for_numpy()
    cD = geom[:, 1:]
    pD = geom[:, 0]
    start = False
    skip_twice = False
    fragment_inds = []
    fragment_ind = 0
    total_index = 0
    lines = p4_input.split("\n")
    for n, l in enumerate(lines):
        if not start and "--" in l:
            start = True
            skip_twice = True
            fragment_inds.append([])
            continue
        if "--" in l:
            skip_twice = True
            fragment_inds.append([])
            fragment_ind += 1
            continue
        if skip_twice:
            skip_twice = False
            continue
        if "".join(l.split()) == "":
            break
        if start:
            if " X " in l:
                continue
            else:
                fragment_inds[fragment_ind].append(total_index)
                total_index += 1
    ma, mb = fragment_inds[0], fragment_inds[1]

    total_charge = 0
    for i in mol.fragment_charges:
        total_charge += i
    charges = np.array(
        [
            [
                int(total_charge),
                int(mol.PYmultiplicity),
            ],
            [int(mol.fragment_charges[0]), (mol.fragment_multiplicities[0])],
            [int(mol.fragment_charges[1]), (mol.fragment_multiplicities[1])],
        ]
    )
    qcel.models.Molecule.from_data
    return p4_input, geom, pD, cD, ma, mb, charges


def remove_extra_wb(line: str):
    """
    Removes extra whitespace in a string for better splitting.
    """
    line = (
        line.replace("    ", " ")
        .replace("   ", " ")
        .replace("  ", " ")
        .replace("  ", " ")
        .replace("\n ", "\n")
    )
    return line


def psi4_input_to_geom_monABs_charges(psi4_input, output_units="angstrom"):
    mol = qcel.models.Molecule.from_data(psi4_input)
    geom = mol.geometry
    if output_units == "angstrom":
        geom = geom * qcel.constants.conversion_factor("bohr", "angstrom")
    Z = mol.atomic_numbers
    geom = np.hstack((Z.reshape(-1, 1), geom))
    monAs = np.array(mol.fragments[0])
    monBs = np.array(mol.fragments[1])
    charges = np.array(
        [
            [int(mol.molecular_charge), int(mol.molecular_multiplicity)],
            [int(mol.fragment_charges[0]), int(mol.fragment_multiplicities[0])],
            [int(mol.fragment_charges[1]), int(mol.fragment_multiplicities[1])],
        ]
    )
    return geom, monAs, monBs, charges


def combine_geometries(pA, pB, geomA, geomB):
    """
    combine_geometries takes in two geometries and combines them
    """
    pA = pA.reshape(-1, 1)
    pB = pB.reshape(-1, 1)
    pA = np.hstack((pA, geomA))
    pB = np.hstack((pB, geomB))
    return np.vstack((pA, pB))


def read_psi4_input_file_molecule(input_path):
    with open(input_path, "r") as f:
        lines = f.readlines()
    geom = []
    start = False
    for n, l in enumerate(lines):
        if "mol" in l:
            start = True
        elif "}" in l:
            break
        elif start:
            geom.append(l)
    geom = "".join(geom)
    geom, monAs, monBs, charges = psi4_input_to_geom_monABs_charges(geom)
    return geom, monAs, monBs, charges


def read_psi4_input_molecule(file):
    start_linenumber = subprocess.run(
        f"grep -n 'molecule' {file} | cut -d: -f1",
        shell=True,
        check=True,
        capture_output=True,
    )
    # start_linenumber = subprocess.run(f"grep -n 'molecule {{' {file} | cut -d: -f1", shell=True, check=True, capture_output=True)
    start_linenumber = int(start_linenumber.stdout.decode("utf-8").strip())
    end_linenumber = subprocess.run(
        f"grep -n '}}' {file} | cut -d: -f1",
        shell=True,
        check=True,
        capture_output=True,
    )
    end_linenumber = end_linenumber.stdout.decode("utf-8").strip().split("\n")
    for i in range(len(end_linenumber)):
        end_linenumber[i] = int(end_linenumber[i])
        if end_linenumber[i] > start_linenumber:
            end_linenumber = end_linenumber[i]
            break
    cmd = f"sed -n '{start_linenumber + 1},{end_linenumber}p' {file}"
    out = subprocess.run(cmd, shell=True, check=True, capture_output=True)
    molecule = out.stdout.decode("utf-8", "ignore").strip().split("\n")
    if molecule[-1] == "}":
        molecule.pop()
    elif "}" in molecule[-1]:
        molecule[-1] = molecule[-1].replace("}", "")
    molecule = "\n".join(molecule)
    qcel_molecule = qcel.models.Molecule.from_data(molecule)
    return qcel_molecule


def read_psi4_input_molecule_to_df_monomer(file):
    qc_mol = read_psi4_input_molecule(file)
    geom = qc_mol.geometry
    Z = qc_mol.atomic_numbers
    charges = [int(qc_mol.molecular_charge), qc_mol.molecular_multiplicity]
    return geom, Z, charges


def read_psi4_input_molecule_to_df_dimer(file, verbose=False):
    qc_mol = read_psi4_input_molecule(file)
    geom = qc_mol.geometry
    Z = qc_mol.atomic_numbers
    charges = np.array(
        [
            [int(qc_mol.molecular_charge), qc_mol.molecular_multiplicity],
            [int(qc_mol.fragment_charges[0]), qc_mol.fragment_multiplicities[0]],
            [int(qc_mol.fragment_charges[1]), qc_mol.fragment_multiplicities[1]],
        ]
    )
    geom = np.hstack((Z.reshape(-1, 1), geom))
    monA = qc_mol.fragments[0]
    monB = qc_mol.fragments[1]
    if verbose:
        print_cartesians(geom, symbols=True)
        print(charges)
        print(monA, monB)
    return geom, Z, charges, monA, monB


def geom_bohr_to_ang(geom):
    geom[:, 1:] *= qcel.constants.conversion_factor("bohr", "angstrom")
    return geom


def read_psi4_input_molecule_to_df(monA_p, monB_p=None):
    if not os.path.exists(monA_p):
        # print(f"{monA_p = } does not exist")
        return None, None, None, None
    if monB_p:
        if not os.path.exists(monB_p):
            # print(f"{monB_p = } does not exist")
            return None, None, None, None
        gA, pA, cA = read_psi4_input_molecule_to_df_monomer(monA_p)
        gB, pB, cB = read_psi4_input_molecule_to_df_monomer(monB_p)
        geom = combine_geometries(pA, pB, gA, gB)
        c = [[cA[0] + cB[0], 1], cA, cB]
        monA = [i for i in range(len(pA))]
        monB = [i for i in range(len(pA), len(pA) + len(pB))]
    else:
        geom, _, c, monA, monB = read_psi4_input_molecule_to_df_dimer(monA_p)
    return geom, monA, monB, c


def closest_intermolecular_contact_dimer(geom, monAs, monBs):
    monA = geom[monAs]
    monB = geom[monBs]
    monA = monA[:, 1:]
    monB = monB[:, 1:]

    # Expand dimensions of monA and monB to enable broadcasting
    monA_exp = monA[:, np.newaxis, :]
    monB_exp = monB[np.newaxis, :, :]

    # Calculate pairwise distances using broadcasting and then np.linalg.norm
    dists = np.linalg.norm(monA_exp - monB_exp, axis=2)

    # Find the minimum distance
    min_dist = np.min(dists)
    return min_dist


def mol_to_pdb_for_pymol_visualization_energy(
    mol,
    pairs,
    output_pdb_path,
    create_pml_script=False,
    execute_pml_script=False,
    bounds=None,
    monomers=False,
    bg_color=None,
):
    geom, pD, cD, ma, mb, charges = mol_to_pos_carts_ma_mb(mol)
    pairs_A = np.sum(pairs, axis=1) / 2
    pairs_B = np.sum(pairs, axis=0) / 2
    atom_energies = np.concatenate([pairs_A, pairs_B])
    with open(output_pdb_path, "w") as f:
        for n, r, a in zip(range(len(geom)), geom, atom_energies):
            atom_type = qcel.periodictable.to_E(r[0])
            x_coord, y_coord, z_coord = r[1:]
            f.write(
                f"HETATM{n+1:>5} {atom_type:<2}   001 A   1{x_coord:>12.3f}{y_coord:>8.3f}{z_coord:>8.3f}  1.00{a:>6.2f}{atom_type:>12}\n"
            )
    if monomers:
        with open(output_pdb_path.replace(".pdb", "_monA.pdb"), "w") as f:
            for n, r, a in zip(range(len(ma)), geom[ma], atom_energies[ma]):
                atom_type = qcel.periodictable.to_E(r[0])
                x_coord, y_coord, z_coord = r[1:]
                f.write(
                    f"HETATM{n+1:>5} {atom_type:<2}   001 A   1{x_coord:>12.3f}{y_coord:>8.3f}{z_coord:>8.3f}  1.00{a:>6.2f}{atom_type:>12}\n"
                )
        with open(output_pdb_path.replace(".pdb", "_monB.pdb"), "w") as f:
            for n, r, a in zip(range(len(mb)), geom[mb], atom_energies[mb]):
                atom_type = qcel.periodictable.to_E(r[0])
                x_coord, y_coord, z_coord = r[1:]
                f.write(
                    f"HETATM{n+1:>5} {atom_type:<2}   001 A   1{x_coord:>12.3f}{y_coord:>8.3f}{z_coord:>8.3f}  1.00{a:>6.2f}{atom_type:>12}\n"
                )
    pml_script_path = output_pdb_path.replace(".pdb", ".pml")
    if create_pml_script:
        if bounds is None:
            bounds = [np.min(atom_energies), np.max(atom_energies)]
            # get indices of min and max

        if bg_color is not None:
            bg_color = f"bg {bg_color}"
        else:
            bg_color = ""
        pml_script = f"""
load {output_pdb_path}
color grey, elem c
label all, elem
show spheres,*
set sphere_scale, 0.20
show stick,*
set_bond stick_radius, 0.15, v.
set stick_h_scale,1
zoom center,6
center v.
rotate x, 25
set ray_opaque_background, off
set opaque_background, off
set label_color, black
label all, "%s,%.2f" % (elem, b)
{bg_color}

spectrum b, red_white_blue, minimum = {bounds[0]}, maximum = {bounds[1]}

ray 2000,2000
png {output_pdb_path.replace(".pdb", ".png")}, dpi=400
"""
        with open(pml_script_path, "w") as f:
            f.write(pml_script)
    if execute_pml_script:
        os.system(f"pymol -c {pml_script_path}")
    return
